#include "psublast/libc.h"
#include "sim4.h"
#include "psublast/psublast.h"
#include "splice.h"


extern int encoding[NACHARS];

signal_t gt = {{0, 0, 0, 2},{0, 0, 0, 2},{2, 2, 2, 5},{0, 0, 0, 2}};
signal_t ct = {{0, 0, 0, 2},{2, 2, 2, 5},{0, 0, 0, 2},{0, 0, 0, 2}};
signal_t ag = {{2, 2, 5, 2},{0, 0, 2, 0},{0, 0, 2, 0},{0, 0, 2, 0}};
signal_t ac = {{2, 5, 2, 2},{0, 2, 0, 0},{0, 2, 0, 0},{0, 2, 0, 0}};


#ifdef DEBUG
static void print_splice(splice_t *);
#endif

static splice_t *new_splice(char,int,int,int,int,int);
static void splice_donor(uchar *xseq, uchar *yseq, int M, int N, int *gt_score,
                         int *ct_score, int **max_Gf, int **max_Cf,
                         int **start_Gi, int **start_Ci);
static void splice_acceptor(uchar *xseq, uchar *yseq, int M, int N,
                         int *ag_score, int *ac_score, int **max_Gb,  
                         int **max_Cb, int **end_Gi, int **end_Ci);


void splice(uchar *in_seqx, int ls, int us, int le, int ue,
            uchar *in_seqy, int ys, int ye, splice_t **gcell, splice_t **ccell)
{
   int p, q, *gtscore, *ctscore, *agscore, *acscore;
   int i, tmp;
   int maxCscore, maxGscore, Gxs, Gxe, Gy, Cxs, Cxe, Cy, keep_Ci, keep_Gi;
   int *max_Cf, *max_Gf, *max_Cb, *max_Gb;
   int *start_Gi, *start_Ci, *end_Gi, *end_Ci;
   uchar *s;
   
   s = in_seqx+ls-1;
   gtscore = (int *)ckalloc((us-ls+2)*sizeof(int));
   ctscore = (int *)ckalloc((us-ls+2)*sizeof(int));
   for (p=0; p<=us-ls+1; p++) gtscore[p] = ctscore[p] = 0;
   
   for (p=0; p<=us-ls+1; p++, s++) {
     gtscore[p] += gt[encoding[*s]][encoding[*(s+1)]];
     ctscore[p] += ct[encoding[*s]][encoding[*(s+1)]];
   }
   
   s = in_seqx+ue-1;
   agscore = (int *)ckalloc((ue-le+2)*sizeof(int));
   acscore = (int *)ckalloc((ue-le+2)*sizeof(int));  
   for (q=0; q<=ue-le+1; q++) agscore[q] = acscore[q] = 0;
   
   for (q=ue-le+1; q>=0; q--, s--) {
     agscore[q] += ag[encoding[*(s-1)]][encoding[*s]];
     acscore[q] += ac[encoding[*(s-1)]][encoding[*s]];
   }
   
   splice_donor(in_seqx+ls-1, in_seqy+ys-1, us-ls+1, ye-ys+1, 
                gtscore, ctscore, &max_Gf, &max_Cf, &start_Gi, &start_Ci);

   splice_acceptor(in_seqx+le-1, in_seqy+ys-1, ue-le+1, ye-ys+1, 
                   agscore, acscore, &max_Gb, &max_Cb, &end_Gi, &end_Ci);

   free(gtscore); free(agscore);
   free(ctscore); free(acscore);

   maxCscore = -999999; maxGscore = -999999;
   Gxs = Gxe = Gy = Cxs = Cxe = Cy = -1;
   for (i=0; i<=ye-ys+1; i++) {
     if ((tmp=max_Gf[i]+max_Gb[i])>maxGscore) {
         maxGscore = tmp;
         /* save (i, start_Gi[i], end_Gi[i]); */
         Gxs = ls+start_Gi[i]-1; Gxe = le+end_Gi[i]-1; Gy = ys+i-1;
         keep_Gi = i;
     }  
     if ((tmp=max_Cf[i]+max_Cb[i])>maxCscore) {
         maxCscore = tmp;
         /* save (i, start_Ci[i], end_Ci[i]); */
         Cxs = ls+start_Ci[i]-1; Cxe = le+end_Ci[i]-1; Cy = ys+i-1;
         keep_Ci = i;
     }
   }

   free(max_Gf); free(max_Gb); free(start_Gi); free(end_Gi);
   free(max_Cf); free(max_Cb); free(start_Ci); free(end_Ci);

   *gcell = new_splice('G', Gxs, Gxe, Gy, Gy+1, maxGscore);
   *ccell = new_splice('C', Cxs, Cxe, Cy, Cy+1, maxCscore);

#ifdef DEBUG
   print_splice(*gcell); print_splice(*ccell);
#endif

   return;
}

static splice_t *new_splice(char c, int xs, int xe, int ys, int ye, int score)
{
   splice_t *sp = ckalloc(sizeof(splice_t));
   
   sp->type = c; sp->xs = xs; sp->xe = xe; 
   sp->ys = ys; sp->ye = ye; sp->score = score;

   return sp; 
}

#ifdef DEBUG
static void print_splice(splice_t *g)
{
    printf("Type: %c  sx: %d  se: %d  ys: %d  score: %d\n",
            g->type, g->xs, g->xe, g->ys, g->score);

}
#endif

static void splice_donor(uchar *xseq, uchar *yseq, int M, int N, int *gt_score,
                         int *ct_score, int **max_Gf, int **max_Cf, 
                         int **start_Gi, int **start_Ci)
{
   int *CCf, *mG, *mC, *sC, *sG, *X;
   int i, j, tmp, ss, ssx, cx, c;
   uchar *s, *t;

   CCf = (int *)ckalloc((M+1)*sizeof(int));
   X = (int *)ckalloc((M+1)*sizeof(int));
   mC = *max_Cf = (int *)ckalloc((N+1)*sizeof(int));
   mG = *max_Gf = (int *)ckalloc((N+1)*sizeof(int));
   sG = *start_Gi = (int *)ckalloc((N+1)*sizeof(int));
   sC = *start_Ci = (int *)ckalloc((N+1)*sizeof(int));

   t = yseq; X[0] = CCf[0] = 0;
   for (j=1; j<=M; j++) { CCf[j] = j; X[j] = 0; } 
   
   mG[0] = mC[0] = -999999;
   for (j=0; j<=M; j++) {
      if ((gt_score[j])>mG[0]) { mG[0] = gt_score[j]; sG[0] = j; }
      if ((ct_score[j])>mC[0]) { mC[0] = ct_score[j]; sC[0] = j; } 
   }
   
   for (i=1; i<=N; i++, t++) {
     s = xseq;
     ss = CCf[0]; ssx = X[0];
     c = ++CCf[0]; cx = X[0];
     for (j=1; j<=M; j++, s++) {
          tmp=min(min(CCf[j]+1, ss+(*t!=*s)),c+1);
          if (tmp==c+1);
          else if (tmp==CCf[j]) cx = X[j];
          else cx = ssx + (*t==*s);
          c = tmp; ss = CCf[j]; CCf[j] = c; ssx = X[j]; X[j] = cx;
     }
   
     /* compute max_Gf and max_Cf */
     mG[i] = mC[i] = -999999; 
     for (j=0; j<=M; j++) {
        assert(X[j]+CCf[j]!=0);
        tmp = (int)(12*X[j]/(X[j]+CCf[j])); 
        if ((tmp+gt_score[j])>mG[i]) {
            mG[i] = tmp+gt_score[j]; sG[i] = j;
        }
        if ((tmp+ct_score[j])>mC[i]) {
            mC[i] = tmp+ct_score[j]; sC[i] = j;
        }
     } 
   } 
   free(CCf); free(X); 
}  
     
static void splice_acceptor(uchar *xseq, uchar *yseq, int M, int N, 
                            int *ag_score, int *ac_score, int **max_Gb, 
                            int **max_Cb, int **end_Gi, int **end_Ci)
{
   int *CCb, *X, *mC, *mG, *eC, *eG;
   int tmp, i, j, ss, ssx, cx, c;
   uchar *t, *s;

   CCb = (int *)ckalloc((M+1)*sizeof(int));
   X = (int *)ckalloc((M+1)*sizeof(int));
   mC = *max_Cb = (int *)ckalloc((N+1)*sizeof(int));
   mG = *max_Gb = (int *)ckalloc((N+1)*sizeof(int));
   eG = *end_Gi = (int *)ckalloc((N+1)*sizeof(int));
   eC = *end_Ci = (int *)ckalloc((N+1)*sizeof(int));
   
   t = yseq+N-1; CCb[M] = X[M] = 0;
   for (j=M-1; j>=0; j--) { CCb[j] = M-j; X[j] = 0; }
   
   mG[N] = mC[N] = -999999;
   for (j=M; j>=0; j--) {
      if ((ag_score[j])>mG[N]) { mG[N] = ag_score[j]; eG[N] = j+1; }
      if ((ac_score[j])>mC[N]) { mC[N] = ac_score[j]; eC[N] = j+1; }
   }
   
   for (i=N-1; i>=0; i--, t--) {
     s = xseq+M-1; 
     ss = CCb[M]; ssx = X[M];
     c = ++CCb[M]; cx = X[M];
     for (j=M-1; j>=0; j--, s--) {
          tmp=min(min(CCb[j]+1, ss+(*t!=*s)),c+1);
          if (tmp==c+1) ; 
          else if (tmp==CCb[j]) cx = X[j];
          else cx = ssx + (*t==*s);
          c = tmp; ss = CCb[j]; CCb[j] = c; ssx = X[j]; X[j] = cx;
     }
     
     /* compute max_Gb and max_Cb */
     mG[i] = -999999; mC[i] = -999999;
     for (j=M; j>=0; j--) {
        assert(CCb[j]+X[j]!=0);
        tmp = (int)(12*X[j]/(CCb[j]+X[j]));
        if ((tmp+ag_score[j])>mG[i]) {
            mG[i] = tmp+ag_score[j]; eG[i] = j+1;
        }
        if ((tmp+ac_score[j])>mC[i]) {
            mC[i] = tmp+ac_score[j]; eC[i] = j+1;
        }
     } 
   } 
   free(CCb); free(X); 
}
